@@ -15,7 +15,7 @@ our @ISA = qw(Exporter);
 
 our @EXPORT_OK = qw( html_to_muse html_file_to_muse );
 
-our $VERSION = '0.25';
+our $VERSION = '0.27';
 
 =encoding utf8
 
@@ -42,23 +42,24 @@ use IO::HTML qw/html_file/;
 use HTML::PullParser;
 
 my %preserved = (
-		 "em" => ["<em>", "</em>"],
-		 "i"  => ["<em>", "</em>"],
-		 "u"  => ["<em>", "</em>"],
-		 "strong" => ["<strong>", "</strong>"],
-		 "b"      => ["<strong>", "</strong>"],
+		 "em" => [["<em>"], ["</em>"]],
+		 "i"  => [["<em>"], ["</em>"]],
+		 "u"  => [["<em>"], ["</em>"]],
+		 "strong" => [["<strong>"], ["</strong>"]],
+		 "b"      => [["<strong>"], ["</strong>"]],
 		 "blockquote" => ["\n<quote>\n", "\n</quote>"],
 		 "ol" => ["\n\n", "\n\n"],
 		 "ul" => ["\n\n", "\n\n"],
 		 "li" => { ol => [ " 1. ", "\n\n"],
 			   ul => [ " - ", "\n\n"],
 			 },
-		 "code" => ["<code>", "</code>"],
-		 "a" => ["[[", "]]"],
+		 "code" => [["<code>"], ["</code>"]],
+		 "a" => [[ "[[" ] , [ "]]" ]],
 		 "pre" => [ "\n<example>\n", "\n</example>\n" ],
-		 "tr" => ["\n", "\n"],
-		 "td" => [" | ", " " ],
-		 "th" => [ " || ", " " ],
+		 table => ["\n\n", "\n\n"],
+		 "tr" => ["\n ", "" ],
+		 "td" => [[" "], [" | "] ],
+		 "th" => [[ " "], [" || "] ],
 		 "dd" => ["\n\n", "\n\n"],
 		 "dt" => ["\n***** ", "\n\n" ],
 		 "h1" => ["\n* ", "\n\n"],
@@ -67,10 +68,10 @@ my %preserved = (
 		 "h4" => ["\n*** ", "\n\n"],
 		 "h5" => ["\n**** ", "\n\n"],
 		 "h6" => ["\n***** ", "\n\n"],
-		 "sup" => ["<sup>", "</sup>" ],
-		 "sub" => ["<sub>", "</sub>" ],
-		 "strike" => ["<del>", "</del>"],
-		 "del" => ["<del>", "</del>"],
+		 "sup" => [["<sup>"], ["</sup>"]],
+		 "sub" => [["<sub>"], ["</sub>"]],
+		 "strike" => [["<del>"], ["</del>"]],
+		 "del" => [["<del>"], ["</del>"]],
 		 "p" => ["\n\n", "\n\n"],
 		 "br" => ["\n\n", "\n\n"], # if you're asking why, a
                                            # lot of pages use the br
@@ -179,7 +180,7 @@ sub _html_to_muse {
       if ((defined $tag) &&
 	  ($tag eq 'a') &&
 	  (my $href =  $attr->{href})) {
-	push @textstack, $href, "][";
+	push @textstack, [ $href, "][" ];
       }
     }
 
@@ -230,15 +231,44 @@ sub _html_to_muse {
         }
       }
       $line =~ s/\x{a0}/ /gs;
+      # remove leading spaces from these tags
       if ($current =~ m/^(h[1-6]|li|ul|ol|p|div)$/) {
         $line =~ s/^\s+//gms;
       }
-      push @textstack, $line;
+      if ($current ne 'pre') {
+        push @textstack, [ $line ];
+      }
+      else {
+        push @textstack, $line;
+      }
     } else {
       warn "which type? $type??\n"
     }
   }
-  my $parsed = join("", @textstack);
+  my @current_text;
+  my @processed;
+  while (@textstack) {
+    my $text = shift(@textstack);
+    if (ref($text)) {
+      push @current_text, @$text;
+    }
+    else {
+      push @processed, _merge_text_lines(\@current_text);
+      push @processed, $text;
+    }
+  }
+  push @processed, _merge_text_lines(\@current_text);
+  my $full = join("", @processed);
+  $full =~ s/\n\n\n+/\n\n/gs;
+  return $full;
+}
+
+sub _cleanup_text_block {
+  my $parsed = shift;
+  return '' unless defined $parsed;
+  # here we are inside a single text block.
+  $parsed =~ s/\s+/ /gs;
+  # print "<<<$parsed>>>\n";
   # clean the footnotes.
   $parsed =~ s!\[
 	       \[
@@ -264,9 +294,12 @@ sub _html_to_muse {
     $parsed =~ s!(<[^/]*?>)( +)!$2$1!g;
     $recursion++;
   }
-  $parsed =~ s/(?<=\S) +(?=\S)/ /g;
-  $parsed =~ s/ +$//gms;
-  $parsed =~ s/\n\n\n+/\n\n/gs;
+  # empty links artefacts.
+  $parsed =~ s/\[\[\]\]//g;
+  $parsed =~ s/\s+/ /gs;
+  $parsed =~ s/\A\s+//;
+  $parsed =~ s/\s+\z//;
+  # print ">>>$parsed<<<\n";
   return $parsed;
 }
 
@@ -308,6 +341,14 @@ sub _pars_process_attr {
   return $tag;
 }
 
+sub _merge_text_lines {
+  my $lines = shift;
+  return '' unless @$lines;
+  my $text = join ('', @$lines);
+  @$lines = ();
+  return _cleanup_text_block($text);
+}
+
 1;
 
 
